from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from datetime import datetime
from typing import Any
import dataclasses

from restate import RunOptions, SdkInternalBaseException
from restate.ext.pydantic._utils import current_state
from restate.extensions import current_context
from restate.ext.turnstile import Turnstile

from pydantic_ai.agent.abstract import EventStreamHandler
from pydantic_ai.exceptions import UserError
from pydantic_ai.messages import ModelMessage, ModelResponse, ModelResponseStreamEvent
from pydantic_ai.models import Model, ModelRequestParameters, StreamedResponse
from pydantic_ai.models.wrapper import WrapperModel
from pydantic_ai.settings import ModelSettings
from pydantic_ai.tools import AgentDepsT, RunContext
from pydantic_ai.usage import RequestUsage


from ._serde import PydanticTypeAdapter

MODEL_RESPONSE_SERDE = PydanticTypeAdapter(ModelResponse)


class RestateStreamedResponse(StreamedResponse):
    def __init__(self, model_request_parameters: ModelRequestParameters, response: ModelResponse):
        super().__init__(model_request_parameters)
        self.response = response

    async def _get_event_iterator(self) -> AsyncIterator[ModelResponseStreamEvent]:
        return
        # noinspection PyUnreachableCode
        yield

    def get(self) -> ModelResponse:
        return self.response

    def usage(self) -> RequestUsage:
        return self.response.usage  # pragma: no cover

    @property
    def model_name(self) -> str:
        return self.response.model_name or ""  # pragma: no cover

    @property
    def provider_name(self) -> str:
        return self.response.provider_name or ""  # pragma: no cover

    @property
    def timestamp(self) -> datetime:
        return self.response.timestamp  # pragma: no cover

    @property
    def provider_url(self) -> str | None:
        return None


class RestateModelWrapper(WrapperModel):
    def __init__(
        self,
        wrapped: Model,
        run_options: RunOptions,
        event_stream_handler: EventStreamHandler[AgentDepsT] | None = None,
    ):
        super().__init__(wrapped)

        self._options = dataclasses.replace(run_options, serde=MODEL_RESPONSE_SERDE)
        self._event_stream_handler = event_stream_handler

    async def request(self, *args: Any, **kwargs: Any) -> ModelResponse:
        context = current_context()
        if context is None:
            raise UserError(
                "A model cannot be used without a Restate context. Make sure to run it within an agent or a run context."
            )
        try:
            res = await context.run_typed("Model call", self.wrapped.request, self._options, *args, **kwargs)
            ids = [c.tool_call_id for c in res.tool_calls]
            current_state().turnstile = Turnstile(ids)
            return res
        except SdkInternalBaseException as e:
            raise Exception("Internal error during model call") from e

    @asynccontextmanager
    async def request_stream(
        self,
        messages: list[ModelMessage],
        model_settings: ModelSettings | None,
        model_request_parameters: ModelRequestParameters,
        run_context: RunContext[AgentDepsT] | None = None,
    ) -> AsyncIterator[StreamedResponse]:
        if run_context is None:
            raise UserError(
                "A model cannot be used with `pydantic_ai.direct.model_request_stream()` as it requires a `run_context`. Set an `event_stream_handler` on the agent and use `agent.run()` instead."
            )

        fn = self._event_stream_handler
        assert fn is not None

        async def request_stream_run():
            async with self.wrapped.request_stream(
                messages,
                model_settings,
                model_request_parameters,
                run_context,
            ) as streamed_response:
                await fn(run_context, streamed_response)  # type: ignore[arg-type]

                async for _ in streamed_response:
                    pass
            return streamed_response.get()

        try:
            response = await self._context.run_typed("Model stream call", request_stream_run, self._options)
            yield RestateStreamedResponse(model_request_parameters, response)
        except SdkInternalBaseException as e:
            raise Exception("Internal error during model stream call") from e



from __future__ import annotations

from collections.abc import AsyncIterable, AsyncIterator, Iterator, Sequence
from contextlib import AbstractAsyncContextManager, asynccontextmanager, contextmanager
from typing import Any, overload

from restate import RunOptions, TerminalError
from restate.ext.pydantic._utils import state_context
from restate.extensions import current_context

from pydantic_ai import models
from pydantic_ai._run_context import AgentDepsT
from pydantic_ai.agent.abstract import AbstractAgent, EventStreamHandler, RunOutputDataT, Instructions
from pydantic_ai.agent.wrapper import WrapperAgent
from pydantic_ai.builtin_tools import AbstractBuiltinTool
from pydantic_ai.exceptions import UserError
from pydantic_ai.messages import AgentStreamEvent, ModelMessage, UserContent
from pydantic_ai.models import Model
from pydantic_ai.output import OutputDataT, OutputSpec
from pydantic_ai.result import StreamedRunResult
from pydantic_ai.run import AgentRunResult
from pydantic_ai.settings import ModelSettings
from pydantic_ai.tools import DeferredToolResults, RunContext, BuiltinToolFunc
from pydantic_ai.toolsets.abstract import AbstractToolset
from pydantic_ai.toolsets.function import FunctionToolset
from pydantic_ai.usage import RunUsage, UsageLimits
from ._model import RestateModelWrapper
from ._toolset import RestateContextRunToolSet


class RestateAgent(WrapperAgent[AgentDepsT, OutputDataT]):
    """An agent that integrates with Restate framework for building resilient applications.

    This agent wraps an existing agent with Restate context capabilities, providing
    automatic retries and durable execution for all operations. By default, tool calls
    are automatically wrapped with Restate's execution model.

    Example:
       ...

       weather = restate.Service('weather')

       agent = RestateAgent(weather_agent)

       @weather.handler()
       async def get_weather(ctx: restate.Context, city: str):
            result = await agent.run(f'What is the weather in {city}?')
            return result.output
       ...

    For advanced scenarios, you can disable automatic tool wrapping by setting
    `disable_auto_wrapping_tools=True`. This allows direct usage of Restate context
    within your tools for features like RPC calls, timers, and multi-step operations.

    When automatic wrapping is disabled, function tools will NOT be automatically executed
    within Restate's `ctx.run()` context, giving you full control over how the
    Restate context is used within your tool implementations.
    But model calls, and MCP tool calls will still be automatically wrapped.

    Example:
       ...

       @dataclass
       WeatherDeps:
            ...
            restate_context: Context

       weather_agent = Agent(..., deps_type=WeatherDeps, ...)

       @weather_agent.tool
       async def get_lat_lng(ctx: RunContext[WeatherDeps], location_description: str) -> LatLng:
            restate_context = ctx.deps.restate_context
            lat = await restate_context.run(...) # <---- note the direct usage of the restate context
            lng = await restate_context.run(...)
            return LatLng(lat, lng)


       agent = RestateAgent(weather_agent)

       weather = restate.Service('weather')

       @weather.handler()
       async def get_weather(ctx: restate.Context, city: str):
            result = await agent.run(f'What is the weather in {city}?', deps=WeatherDeps(restate_context=ctx, ...))
            return result.output
       ...

    """

    def __init__(
        self,
        wrapped: AbstractAgent[AgentDepsT, OutputDataT],
        *,
        event_stream_handler: EventStreamHandler[AgentDepsT] | None = None,
        disable_auto_wrapping_tools: bool = False,
        run_options: RunOptions | None = None,
    ):
        super().__init__(wrapped)
        if not isinstance(wrapped.model, Model):
            raise TerminalError(
                "An agent needs to have a `model` in order to be used with Restate, it cannot be set at agent run time."
            )

        self._event_stream_handler = event_stream_handler
        self._disable_auto_wrapping_tools = disable_auto_wrapping_tools

        if run_options is None:
            run_options = RunOptions(max_attempts=3)

        self._model = RestateModelWrapper(wrapped.model, run_options, event_stream_handler=event_stream_handler)

        def set_context(toolset: AbstractToolset[AgentDepsT]) -> AbstractToolset[AgentDepsT]:
            """Set the Restate context for the toolset, wrapping tools if needed."""
            if isinstance(toolset, FunctionToolset) and not disable_auto_wrapping_tools:
                return RestateContextRunToolSet(toolset, run_options)
            try:
                from pydantic_ai.mcp import MCPServer

                from ._toolset import RestateMCPServer
            except ImportError:
                pass
            else:
                if isinstance(toolset, MCPServer):
                    return RestateMCPServer(toolset, run_options)

            return toolset

        self._toolsets = [toolset.visit_and_replace(set_context) for toolset in wrapped.toolsets]

    @contextmanager
    def _restate_overrides(self) -> Iterator[None]:
        with (
            super().override(model=self._model, toolsets=self._toolsets, tools=[]),
        ):
            with state_context():
                yield

    @property
    def model(self) -> models.Model | models.KnownModelName | str | None:
        return self._model

    @property
    def event_stream_handler(self) -> EventStreamHandler[AgentDepsT] | None:
        handler = self._event_stream_handler or super().event_stream_handler
        if handler is None:
            return None
        if self._disable_auto_wrapping_tools:
            return handler
        return self.wrapped_event_stream_handler

    async def wrapped_event_stream_handler(
        self, ctx: RunContext[AgentDepsT], stream: AsyncIterable[AgentStreamEvent]
    ) -> None:
        fn = self._event_stream_handler
        if fn is None:
            return
        context = current_context()
        if context is None:
            raise UserError("No Restate context found for RestateAgent event stream handler.")

        async for event in stream:

            async def single_event():
                yield event

            await context.run_typed("run event", lambda: fn(ctx, single_event()))

    @property
    def toolsets(self) -> Sequence[AbstractToolset[AgentDepsT]]:
        with self._restate_overrides():
            return super().toolsets

    @overload
    async def run(
        self,
        user_prompt: str | Sequence[UserContent] | None = None,
        *,
        output_type: None = None,
        message_history: Sequence[ModelMessage] | None = None,
        deferred_tool_results: DeferredToolResults | None = None,
        model: models.Model | models.KnownModelName | str | None = None,
        instructions: Instructions[AgentDepsT] = None,
        deps: AgentDepsT = None,
        model_settings: ModelSettings | None = None,
        usage_limits: UsageLimits | None = None,
        usage: RunUsage | None = None,
        infer_name: bool = True,
        toolsets: Sequence[AbstractToolset[AgentDepsT]] | None = None,
        builtin_tools: Sequence[AbstractBuiltinTool | BuiltinToolFunc[AgentDepsT]] | None = None,
        event_stream_handler: EventStreamHandler[AgentDepsT] | None = None,
    ) -> AgentRunResult[OutputDataT]: ...

    @overload
    async def run(
        self,
        user_prompt: str | Sequence[UserContent] | None = None,
        *,
        output_type: OutputSpec[RunOutputDataT],
        message_history: Sequence[ModelMessage] | None = None,
        deferred_tool_results: DeferredToolResults | None = None,
        model: models.Model | models.KnownModelName | str | None = None,
        instructions: Instructions[AgentDepsT] = None,
        deps: AgentDepsT = None,
        model_settings: ModelSettings | None = None,
        usage_limits: UsageLimits | None = None,
        usage: RunUsage | None = None,
        infer_name: bool = True,
        toolsets: Sequence[AbstractToolset[AgentDepsT]] | None = None,
        builtin_tools: Sequence[AbstractBuiltinTool | BuiltinToolFunc[AgentDepsT]] | None = None,
        event_stream_handler: EventStreamHandler[AgentDepsT] | None = None,
    ) -> AgentRunResult[RunOutputDataT]: ...

    async def run(
        self,
        user_prompt: str | Sequence[UserContent] | None = None,
        *,
        output_type: OutputSpec[RunOutputDataT] | None = None,
        message_history: Sequence[ModelMessage] | None = None,
        deferred_tool_results: DeferredToolResults | None = None,
        model: models.Model | models.KnownModelName | str | None = None,
        instructions: Instructions[AgentDepsT] = None,
        deps: AgentDepsT = None,
        model_settings: ModelSettings | None = None,
        usage_limits: UsageLimits | None = None,
        usage: RunUsage | None = None,
        infer_name: bool = True,
        toolsets: Sequence[AbstractToolset[AgentDepsT]] | None = None,
        builtin_tools: Sequence[AbstractBuiltinTool | BuiltinToolFunc[AgentDepsT]] | None = None,
        event_stream_handler: EventStreamHandler[AgentDepsT] | None = None,
    ) -> AgentRunResult[Any]:
        """Run the agent with a user prompt in async mode.

        This method builds an internal agent graph (using system prompts, tools and output schemas) and then
        runs the graph to completion. The result of the run is returned.

        Example:
        ```python
        from pydantic_ai import Agent

        agent = Agent('openai:gpt-4o')

        async def main():
            agent_run = await agent.run('What is the capital of France?')
            print(agent_run.output)
            #> The capital of France is Paris.
        ```

        Args:
            user_prompt: User input to start/continue the conversation.
            output_type: Custom output type to use for this run, `output_type` may only be used if the agent has no
                output validators since output validators would expect an argument that matches the agent's output type.
            message_history: History of the conversation so far.
            deferred_tool_results: Optional results for deferred tool calls in the message history.
            model: Optional model to use for this run, required if `model` was not set when creating the agent.
            instructions: Optional additional instructions to use for this run.
            deps: Optional dependencies to use for this run.
            model_settings: Optional settings to use for this model's request.
            usage_limits: Optional limits on model request count or token usage.
            usage: Optional usage to start with, useful for resuming a conversation or agents used in tools.
            infer_name: Whether to try to infer the agent name from the call frame if it's not set.
            toolsets: Optional additional toolsets for this run.
            event_stream_handler: Optional handler for events from the model's streaming response and the agent's execution of tools to use for this run.
            builtin_tools: Optional additional builtin tools for this run.

        Returns:
            The result of the run.
        """
        if model is not None:
            raise TerminalError(
                "An agent needs to have a `model` in order to be used with Restate, it cannot be set at agent run time."
            )
        with self._restate_overrides():
            return await super(WrapperAgent, self).run(
                user_prompt=user_prompt,
                output_type=output_type,
                message_history=message_history,
                instructions=instructions,
                deferred_tool_results=deferred_tool_results,
                model=model,
                deps=deps,
                model_settings=model_settings,
                usage_limits=usage_limits,
                usage=usage,
                infer_name=infer_name,
                toolsets=toolsets,
                builtin_tools=builtin_tools,
                event_stream_handler=event_stream_handler,
            )

    @overload
    def run_stream(
        self,
        user_prompt: str | Sequence[UserContent] | None = None,
        *,
        output_type: None = None,
        message_history: Sequence[ModelMessage] | None = None,
        deferred_tool_results: DeferredToolResults | None = None,
        model: models.Model | models.KnownModelName | str | None = None,
        instructions: Instructions[AgentDepsT] = None,
        deps: AgentDepsT = None,
        model_settings: ModelSettings | None = None,
        usage_limits: UsageLimits | None = None,
        usage: RunUsage | None = None,
        infer_name: bool = True,
        toolsets: Sequence[AbstractToolset[AgentDepsT]] | None = None,
        builtin_tools: Sequence[AbstractBuiltinTool | BuiltinToolFunc[AgentDepsT]] | None = None,
        event_stream_handler: EventStreamHandler[AgentDepsT] | None = None,
    ) -> AbstractAsyncContextManager[StreamedRunResult[AgentDepsT, OutputDataT]]: ...

    @overload
    def run_stream(
        self,
        user_prompt: str | Sequence[UserContent] | None = None,
        *,
        output_type: OutputSpec[RunOutputDataT],
        message_history: Sequence[ModelMessage] | None = None,
        deferred_tool_results: DeferredToolResults | None = None,
        model: models.Model | models.KnownModelName | str | None = None,
        instructions: Instructions[AgentDepsT] = None,
        deps: AgentDepsT = None,
        model_settings: ModelSettings | None = None,
        usage_limits: UsageLimits | None = None,
        usage: RunUsage | None = None,
        infer_name: bool = True,
        toolsets: Sequence[AbstractToolset[AgentDepsT]] | None = None,
        builtin_tools: Sequence[AbstractBuiltinTool | BuiltinToolFunc[AgentDepsT]] | None = None,
        event_stream_handler: EventStreamHandler[AgentDepsT] | None = None,
    ) -> AbstractAsyncContextManager[StreamedRunResult[AgentDepsT, RunOutputDataT]]: ...

    @asynccontextmanager
    async def run_stream(
        self,
        user_prompt: str | Sequence[UserContent] | None = None,
        *,
        output_type: OutputSpec[RunOutputDataT] | None = None,
        message_history: Sequence[ModelMessage] | None = None,
        deferred_tool_results: DeferredToolResults | None = None,
        model: models.Model | models.KnownModelName | str | None = None,
        instructions: Instructions[AgentDepsT] = None,
        deps: AgentDepsT = None,
        model_settings: ModelSettings | None = None,
        usage_limits: UsageLimits | None = None,
        usage: RunUsage | None = None,
        infer_name: bool = True,
        toolsets: Sequence[AbstractToolset[AgentDepsT]] | None = None,
        builtin_tools: Sequence[AbstractBuiltinTool | BuiltinToolFunc[AgentDepsT]] | None = None,
        event_stream_handler: EventStreamHandler[AgentDepsT] | None = None,
        **_deprecated_kwargs: Any,
    ) -> AsyncIterator[StreamedRunResult[AgentDepsT, Any]]:
        """Run the agent with a user prompt in async mode, returning a streamed response.

        Example:
        ```python
        from pydantic_ai import Agent

        agent = Agent('openai:gpt-4o')

        async def main():
            async with agent.run_stream('What is the capital of the UK?') as response:
                print(await response.get_output())
                #> The capital of the UK is London.
        ```

        Args:
            user_prompt: User input to start/continue the conversation.
            output_type: Custom output type to use for this run, `output_type` may only be used if the agent has no
                output validators since output validators would expect an argument that matches the agent's output type.
            message_history: History of the conversation so far.
            deferred_tool_results: Optional results for deferred tool calls in the message history.
            model: Optional model to use for this run, required if `model` was not set when creating the agent.
            instructions: Optional additional instructions to use for this run.
            deps: Optional dependencies to use for this run.
            model_settings: Optional settings to use for this model's request.
            usage_limits: Optional limits on model request count or token usage.
            usage: Optional usage to start with, useful for resuming a conversation or agents used in tools.
            infer_name: Whether to try to infer the agent name from the call frame if it's not set.
            toolsets: Optional additional toolsets for this run.
            builtin_tools: Optional additional builtin tools for this run.
            event_stream_handler: Optional event stream handler to use for this run. It will receive all the events up until the final result is found, which you can then read or stream from inside the context manager.

        Returns:
            The result of the run.
        """
        raise UserError("RestateAgent does not support run_stream currently.")
        yield



#
#  Copyright (c) 2023-2025 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
ADK plugin implementation for restate.
"""

import restate

from datetime import timedelta
from typing import Optional, Any, cast

from google.genai import types

from google.adk.agents import BaseAgent, LlmAgent
from google.adk.agents.callback_context import CallbackContext
from google.adk.agents.invocation_context import InvocationContext
from google.adk.plugins import BasePlugin
from google.adk.tools.base_tool import BaseTool
from google.adk.tools.tool_context import ToolContext
from google.adk.models.llm_request import LlmRequest
from google.adk.models.llm_response import LlmResponse
from google.adk.models import LLMRegistry
from google.adk.models.base_llm import BaseLlm
from google.adk.flows.llm_flows.functions import generate_client_function_call_id
from restate.ext.adk import RestateSessionService

from restate.extensions import current_context

from restate.ext.turnstile import Turnstile
from restate.server_context import clear_extension_data, get_extension_data, set_extension_data


def _create_turnstile(s: LlmResponse) -> Turnstile:
    ids = _get_function_call_ids(s)
    turnstile = Turnstile(ids)
    return turnstile


def _invocation_extension_key(invocation_id: str) -> str:
    return "adk_" + invocation_id


class PluginState:
    def __init__(self, model: BaseLlm):
        self.model = model
        self.turnstile: Turnstile = Turnstile([])

    def __close__(self):
        """Clean up resources."""
        self.turnstile.cancel_all()

    @staticmethod
    def create(invocation_id: str, model: BaseLlm, ctx: restate.Context) -> None:
        extension_key = _invocation_extension_key(invocation_id)
        state = PluginState(model)
        set_extension_data(ctx, extension_key, state)

    @staticmethod
    def clear(invocation_id: str) -> None:
        ctx = current_context()
        if ctx is None:
            raise RuntimeError(
                "No Restate context found, the restate plugin must be used from within a restate handler."
            )
        extension_key = _invocation_extension_key(invocation_id)
        clear_extension_data(ctx, extension_key)

    @staticmethod
    def from_context(invocation_id: str, ctx: restate.Context | None = None) -> "PluginState":
        if ctx is None:
            ctx = current_context()
        if ctx is None:
            raise RuntimeError(
                "No Restate context found, the restate plugin must be used from within a restate handler."
            )
        extension_key = _invocation_extension_key(invocation_id)
        state: PluginState | None = get_extension_data(ctx, extension_key)
        if state is None:
            raise RuntimeError(
                "No RestatePlugin state found, the restate plugin must be used from within a restate handler."
            )
        return state


class RestatePlugin(BasePlugin):
    """A plugin to integrate Restate with the ADK framework."""

    def __init__(self, *, max_model_call_retries: int = 10):
        super().__init__(name="restate_plugin")
        self._max_model_call_retries = max_model_call_retries

    async def before_agent_callback(
        self, *, agent: BaseAgent, callback_context: CallbackContext
    ) -> Optional[types.Content]:
        if not isinstance(agent, LlmAgent):
            raise restate.TerminalError("RestatePlugin only supports LlmAgent agents.")
        ctx = current_context()  # Ensure we have a Restate context
        if ctx is None:
            raise restate.TerminalError(
                """No Restate context found for RestatePlugin.
            Ensure that the agent is invoked within a restate handler and,
            using a ```with restate_overrides(ctx):``` block. around your agent use."""
            )

        model = agent.model if isinstance(agent.model, BaseLlm) else LLMRegistry.new_llm(agent.model)
        PluginState.create(callback_context.invocation_id, model, ctx)
        return None

    async def after_agent_callback(
        self, *, agent: BaseAgent, callback_context: CallbackContext
    ) -> Optional[types.Content]:
        PluginState.clear(callback_context.invocation_id)
        return None

    async def after_run_callback(self, *, invocation_context: InvocationContext) -> None:
        if isinstance(invocation_context.session_service, RestateSessionService):
            restate_session_service = cast(RestateSessionService, invocation_context.session_service)
            await restate_session_service.flush_session_state(invocation_context.session)

    async def before_model_callback(
        self, *, callback_context: CallbackContext, llm_request: LlmRequest
    ) -> Optional[LlmResponse]:
        ctx = current_context()  # Ensure we have a Restate context
        if ctx is None:
            raise restate.TerminalError(
                """No Restate context found for RestatePlugin.
            Ensure that the agent is invoked within a restate handler and,
            using a ```with restate_overrides(ctx):``` block. around your agent use."""
            )
        state = PluginState.from_context(callback_context.invocation_id, ctx)
        model = state.model
        response = await _generate_content_async(ctx, self._max_model_call_retries, model, llm_request)
        turnstile = _create_turnstile(response)
        state.turnstile = turnstile
        return response

    async def before_tool_callback(
        self,
        *,
        tool: BaseTool,
        tool_args: dict[str, Any],
        tool_context: ToolContext,
    ) -> Optional[dict]:
        id = tool_context.function_call_id
        assert id is not None, "Function call ID is required for tool invocation."

        turnstile = PluginState.from_context(tool_context.invocation_id).turnstile
        await turnstile.wait_for(id)
        return None

    async def after_tool_callback(
        self,
        *,
        tool: BaseTool,
        tool_args: dict[str, Any],
        tool_context: ToolContext,
        result: dict,
    ) -> Optional[dict]:
        id = tool_context.function_call_id
        assert id is not None, "Function call ID is required for tool invocation."
        turnstile = PluginState.from_context(tool_context.invocation_id).turnstile
        turnstile.allow_next_after(id)

        return None

    async def on_tool_error_callback(
        self,
        *,
        tool: BaseTool,
        tool_args: dict[str, Any],
        tool_context: ToolContext,
        error: Exception,
    ) -> Optional[dict]:
        id = tool_context.function_call_id
        assert id is not None, "Function call ID is required for tool invocation."
        turnstile = PluginState.from_context(tool_context.invocation_id).turnstile
        turnstile.cancel_all_after(id)
        return None


def _get_function_call_ids(s: LlmResponse) -> list[str]:
    ids = []
    if s.content and s.content.parts:
        for part in s.content.parts:
            if part.function_call:
                if part.function_call.id:
                    ids.append(part.function_call.id)
    return ids


def _generate_client_function_call_id(s: LlmResponse) -> None:
    """Generate client function call IDs for function calls in the LlmResponse.
    It is important for the function call IDs to be stable across retries, as they
    are used to correlate function call results with their invocations.
    """
    if s.content and s.content.parts:
        for part in s.content.parts:
            if part.function_call:
                if not part.function_call.id:
                    id = generate_client_function_call_id()
                    part.function_call.id = id


async def _generate_content_async(
    ctx: restate.Context, max_attempts: int, model: BaseLlm, llm_request: LlmRequest
) -> LlmResponse:
    """Generate content using Restate's context."""

    async def call_llm() -> LlmResponse:
        a_gen = model.generate_content_async(llm_request, stream=False)
        try:
            result = await anext(a_gen)
            _generate_client_function_call_id(result)
            return result
        finally:
            await a_gen.aclose()

    return await ctx.run_typed(
        "call LLM",
        call_llm,
        restate.RunOptions(max_attempts=max_attempts, initial_retry_interval=timedelta(seconds=1)),
    )



#
#  Copyright (c) 2023-2025 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
This module contains the optional OpenAI integration for Restate.
"""

import dataclasses

from agents import (
    Model,
    RunConfig,
    TContext,
    RunResult,
    Agent,
    Runner,
)
from agents.models.multi_provider import MultiProvider
from agents.items import TResponseStreamEvent, ModelResponse
from agents.items import TResponseInputItem
from typing import AsyncIterator

from restate.ext.turnstile import Turnstile
from restate.extensions import current_context
from restate import RunOptions, TerminalError

from .functions import get_function_call_ids, wrap_agent_tools
from .models import LlmRetryOpts, RestateModelResponse, State
from .session import RestateSession


class DurableModelCalls(MultiProvider):
    """
    A Restate model provider that wraps the OpenAI SDK's default MultiProvider.
    """

    def __init__(self, state: State, llm_retry_opts: LlmRetryOpts | None = None):
        super().__init__()
        self.llm_retry_opts = llm_retry_opts
        self.state = state

    def get_model(self, model_name: str | None) -> Model:
        model = super().get_model(model_name or None)
        return RestateModelWrapper(model, self.state, self.llm_retry_opts)


class RestateModelWrapper(Model):
    """
    A wrapper around the OpenAI SDK's Model that persists LLM calls in the Restate journal.
    """

    def __init__(self, model: Model, state: State, llm_retry_opts: LlmRetryOpts | None = None):
        self.model = model
        self.state = state
        self.model_name = "RestateModelWrapper"
        self.llm_retry_opts = llm_retry_opts if llm_retry_opts is not None else LlmRetryOpts()

    async def get_response(self, *args, **kwargs) -> ModelResponse:
        async def call_llm() -> RestateModelResponse:
            resp = await self.model.get_response(*args, **kwargs)
            # convert to pydantic model to be serializable by Restate SDK
            return RestateModelResponse(
                output=resp.output,
                usage=resp.usage,
                response_id=resp.response_id,
            )

        ctx = current_context()
        if ctx is None:
            raise RuntimeError("No current Restate context found, make sure to run inside a Restate handler")
        result = await ctx.run_typed(
            "call LLM",
            call_llm,
            RunOptions(
                max_attempts=self.llm_retry_opts.max_attempts,
                max_duration=self.llm_retry_opts.max_duration,
                initial_retry_interval=self.llm_retry_opts.initial_retry_interval,
                max_retry_interval=self.llm_retry_opts.max_retry_interval,
                retry_interval_factor=self.llm_retry_opts.retry_interval_factor,
            ),
        )
        # collect function call IDs, too
        ids = get_function_call_ids(result.output)
        self.state.turnstile = Turnstile(ids)

        # convert back to original ModelResponse
        return ModelResponse(
            output=result.output,
            usage=result.usage,
            response_id=result.response_id,
        )

    def stream_response(self, *args, **kwargs) -> AsyncIterator[TResponseStreamEvent]:
        raise TerminalError("Streaming is not supported in Restate. Use `get_response` instead.")


class DurableRunner:
    """
    A wrapper around Runner.run that automatically configures RunConfig for Restate contexts.

    This class automatically sets up the appropriate model provider (DurableModelCalls) and
    model settings, taking over any model and model_settings configuration provided in the
    original RunConfig.
    """

    @staticmethod
    async def run(
        starting_agent: Agent[TContext],
        input: str | list[TResponseInputItem],
        *,
        use_restate_session: bool = False,
        **kwargs,
    ) -> RunResult:
        """
        Run an agent with automatic Restate configuration.

        Args:
            use_restate_session: If True, creates a RestateSession for conversation persistence.
                                Requires running within a Restate Virtual Object context.

        Returns:
            The result from Runner.run
        """

        # execution state
        state = State()

        # Set persisting model calls
        llm_retry_opts = kwargs.pop("llm_retry_opts", None)
        run_config = kwargs.pop("run_config", RunConfig())
        run_config = dataclasses.replace(run_config, model_provider=DurableModelCalls(state, llm_retry_opts))

        # Use Restate session if requested, otherwise use provided session
        session = kwargs.pop("session", None)
        if use_restate_session:
            if session is not None:
                raise TerminalError("When use_restate_session is True, session config cannot be provided.")
            session = RestateSession()

        agent = wrap_agent_tools(starting_agent, state)
        try:
            result = await Runner.run(
                starting_agent=agent, input=input, run_config=run_config, session=session, **kwargs
            )
        finally:
            # Flush session items to Restate
            if session is not None and isinstance(session, RestateSession):
                session.flush()

        return result



#
#  Copyright (c) 2023-2025 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
import typing

from .session import RestateSessionService
from .plugin import RestatePlugin
from restate import ObjectContext, Context
from restate.extensions import current_context


def restate_object_context() -> ObjectContext:
    """Get the current Restate ObjectContext."""
    ctx = current_context()
    if ctx is None:
        raise RuntimeError("No Restate context found.")
    return typing.cast(ObjectContext, ctx)


def restate_context() -> Context:
    """Get the current Restate Context."""
    ctx = current_context()
    if ctx is None:
        raise RuntimeError("No Restate context found.")
    return ctx


__all__ = [
    "RestateSessionService",
    "RestatePlugin",
    "restate_object_context",
    "restate_context",
]



import typing

from pydantic import TypeAdapter
from restate.serde import Serde

T = typing.TypeVar("T")


class PydanticTypeAdapter(Serde[T]):
    """A serializer/deserializer for Pydantic models."""

    def __init__(self, model_type: type[T]):
        """Initializes a new instance of the PydanticTypeAdaptorSerde class.

        Args:
            model_type (typing.Type[T]): The Pydantic model type to serialize/deserialize.
        """
        self._model_type = TypeAdapter(model_type)

    def deserialize(self, buf: bytes) -> T | None:
        """Deserializes a bytearray to a Pydantic model.

        Args:
            buf (bytearray): The bytearray to deserialize.

        Returns:
            typing.Optional[T]: The deserialized Pydantic model.
        """
        if not buf:
            return None
        return self._model_type.validate_json(buf.decode("utf-8"))  # raises if invalid

    def serialize(self, obj: T | None) -> bytes:
        """Serializes a Pydantic model to a bytearray.

        Args:
            obj (typing.Optional[T]): The Pydantic model to serialize.

        Returns:
            bytes: The serialized bytearray.
        """
        if obj is None:
            return b""
        tpe = TypeAdapter(type(obj))
        return tpe.dump_json(obj)



from ._agent import RestateAgent
from ._model import RestateModelWrapper
from ._serde import PydanticTypeAdapter
from ._toolset import RestateContextRunToolSet


__all__ = [
    "RestateModelWrapper",
    "RestateAgent",
    "PydanticTypeAdapter",
    "RestateContextRunToolSet",
]



#
#  Copyright (c) 2023-2025 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
ADK session service implementation using Restate Virtual Objects as the backing store.
"""

import restate

from typing import Optional, Any, cast

from google.adk.sessions import Session
from google.adk.sessions.state import State
from google.adk.events.event import Event
from google.adk.sessions.base_session_service import (
    BaseSessionService,
    ListSessionsResponse,
    GetSessionConfig,
)
from restate import TerminalError

from restate.extensions import current_context


class RestateSessionService(BaseSessionService):
    def ctx(self) -> restate.ObjectContext:
        return cast(restate.ObjectContext, current_context())

    async def create_session(
        self,
        *,
        app_name: str,
        user_id: str,
        state: Optional[dict[str, Any]] = None,
        session_id: Optional[str] = None,
    ) -> Session:
        if session_id is None:
            session_id = str(self.ctx().uuid())

        session = await self.ctx().get(f"session_store::{session_id}", type_hint=Session)
        if session is not None:
            raise TerminalError("Session with the given ID already exists.")

        session = Session(
            app_name=app_name,
            user_id=user_id,
            id=session_id,
            state=state or {},
        )

        await self.flush_session_state(session)
        return session

    async def has_session(self, *, session_id: str) -> bool:
        return await self.ctx().get(f"session_store::{session_id}", type_hint=Session) is not None

    async def get_session(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        config: Optional[GetSessionConfig] = None,
    ) -> Optional[Session]:
        # TODO : Handle config options
        return await self.ctx().get(f"session_store::{session_id}", type_hint=Session) or Session(
            app_name=app_name,
            user_id=user_id,
            id=session_id,
        )

    async def list_sessions(self, *, app_name: str, user_id: Optional[str] = None) -> ListSessionsResponse:
        state_keys = await self.ctx().state_keys()
        sessions = []
        for key in state_keys:
            if key.startswith("session_store::"):
                session = await self.ctx().get(key, type_hint=Session)
                if session is not None:
                    sessions.append(session)
        return ListSessionsResponse(sessions=sessions)

    async def delete_session(self, *, app_name: str, user_id: str, session_id: str) -> None:
        self.ctx().clear(f"session_store::{session_id}")

    async def append_event(self, session: Session, event: Event) -> Event:
        """Appends an event to a session object."""
        if event.partial:
            return event
        # For now, we also store temp state
        event = self._trim_temp_delta_state(event)
        self._update_session_state(session, event)
        session.events.append(event)
        return event

    async def flush_session_state(self, session: Session):
        session_to_store = session.model_copy()

        # Remove temporary state keys before storing
        for key in list(session_to_store.state.keys()):
            if key.startswith(State.TEMP_PREFIX):
                session_to_store.state.pop(key)

        # Remove restate-specific context that got added by the plugin before storing
        session_to_store.state.pop("restate_context", None)

        deterministic_session = await self.ctx().run_typed(
            "store session", lambda: session_to_store, restate.RunOptions(type_hint=Session)
        )
        self.ctx().set(f"session_store::{session.id}", deterministic_session)



from contextlib import contextmanager
from contextvars import ContextVar

from restate.ext.turnstile import Turnstile


class State:
    __slots__ = ("turnstile",)

    def __init__(self):
        self.turnstile = Turnstile([])


restate_state_var = ContextVar("restate_state_var", default=State())


def current_state() -> State:
    return restate_state_var.get()


def set_current_state(state: State):
    restate_state_var.set(state)


@contextmanager
def state_context():
    token = restate_state_var.set(State())
    try:
        yield
    finally:
        restate_state_var.reset(token)



#
#  Copyright (c) 2023-2024 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
This module contains the optional OpenAI integration for Restate.
"""

import dataclasses

from agents import (
    Usage,
    AgentsException,
)
from agents.items import TResponseOutputItem
from agents.items import TResponseInputItem
from datetime import timedelta
from typing import Optional
from pydantic import BaseModel

from restate.ext.turnstile import Turnstile
from restate import TerminalError


class State:
    __slots__ = ("turnstile",)

    def __init__(self) -> None:
        self.turnstile = Turnstile([])


@dataclasses.dataclass
class LlmRetryOpts:
    max_attempts: Optional[int] = 10
    """Max number of attempts (including the initial), before giving up.

    When giving up, the LLM call will throw a `TerminalError` wrapping the original error message."""
    max_duration: Optional[timedelta] = None
    """Max duration of retries, before giving up.

    When giving up, the LLM call will throw a `TerminalError` wrapping the original error message."""
    initial_retry_interval: Optional[timedelta] = timedelta(seconds=1)
    """Initial interval for the first retry attempt.
    Retry interval will grow by a factor specified in `retry_interval_factor`.

    If any of the other retry related fields is specified, the default for this field is 50 milliseconds, otherwise restate will fallback to the overall invocation retry policy."""
    max_retry_interval: Optional[timedelta] = None
    """Max interval between retries.
    Retry interval will grow by a factor specified in `retry_interval_factor`.

    The default is 10 seconds."""
    retry_interval_factor: Optional[float] = None
    """Exponentiation factor to use when computing the next retry delay.

    If any of the other retry related fields is specified, the default for this field is `2`, meaning retry interval will double at each attempt, otherwise restate will fallback to the overall invocation retry policy."""


# The OpenAI ModelResponse class is a dataclass with Pydantic fields.
# The Restate SDK cannot serialize this. So we turn the ModelResponse int a Pydantic model.
class RestateModelResponse(BaseModel):
    output: list[TResponseOutputItem]
    """A list of outputs (messages, tool calls, etc) generated by the model"""

    usage: Usage
    """The usage information for the response."""

    response_id: str | None
    """An ID for the response which can be used to refer to the response in subsequent calls to the
    model. Not supported by all model providers.
    If using OpenAI models via the Responses API, this is the `response_id` parameter, and it can
    be passed to `Runner.run`.
    """

    def to_input_items(self) -> list[TResponseInputItem]:
        return [it.model_dump(exclude_unset=True) for it in self.output]  # type: ignore


class AgentsTerminalException(AgentsException, TerminalError):
    """Exception that is both an AgentsException and a restate.TerminalError."""

    def __init__(self, *args: object) -> None:
        super().__init__(*args)



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, replace
from typing import Any, Literal

from restate import RunOptions, SdkInternalBaseException, TerminalError
from restate.exceptions import SdkInternalException
from restate.extensions import current_context

from pydantic_ai import ToolDefinition
from pydantic_ai._run_context import AgentDepsT
from pydantic_ai.exceptions import ApprovalRequired, CallDeferred, ModelRetry, UserError
from pydantic_ai.mcp import MCPServer, ToolResult
from pydantic_ai.tools import RunContext
from pydantic_ai.toolsets.abstract import AbstractToolset, ToolsetTool
from pydantic_ai.toolsets.wrapper import WrapperToolset

from ._serde import PydanticTypeAdapter
from ._utils import current_state


@dataclass
class RestateContextRunResult:
    """A simple wrapper for tool results to be used with Restate's run_typed."""

    kind: Literal["output", "call_deferred", "approval_required", "model_retry"]
    output: Any
    error: str | None = None


CONTEXT_RUN_SERDE = PydanticTypeAdapter(RestateContextRunResult)


@dataclass
class RestateMCPGetToolsContextRunResult:
    """A simple wrapper for tool results to be used with Restate's run_typed."""

    output: dict[str, ToolDefinition]


MCP_GET_TOOLS_SERDE = PydanticTypeAdapter(RestateMCPGetToolsContextRunResult)


@dataclass
class RestateMCPToolRunResult:
    """A simple wrapper for tool results to be used with Restate's run_typed."""

    output: ToolResult


MCP_RUN_SERDE = PydanticTypeAdapter(RestateMCPToolRunResult)


class RestateContextRunToolSet(WrapperToolset[AgentDepsT]):
    """A toolset that automatically wraps tool calls with restate's `ctx.run_typed()`."""

    def __init__(self, wrapped: AbstractToolset[AgentDepsT], run_options: RunOptions):
        super().__init__(wrapped)
        self.options = replace(run_options, serde=CONTEXT_RUN_SERDE)

    async def call_tool(
        self, name: str, tool_args: dict[str, Any], ctx: RunContext[AgentDepsT], tool: ToolsetTool[AgentDepsT]
    ) -> Any:
        async def action() -> RestateContextRunResult:
            try:
                # A tool may raise ModelRetry, CallDeferred, ApprovalRequired, or UserError
                # to signal special conditions to the caller.
                # Since, restate ctx.run() will retry this exception we need to convert these exceptions
                # to a return value and handle them outside of the ctx.run().
                output = await self.wrapped.call_tool(name, tool_args, ctx, tool)
                return RestateContextRunResult(kind="output", output=output, error=None)
            except ModelRetry as e:
                return RestateContextRunResult(kind="model_retry", output=None, error=e.message)
            except CallDeferred:
                return RestateContextRunResult(kind="call_deferred", output=None, error=None)
            except ApprovalRequired:
                return RestateContextRunResult(kind="approval_required", output=None, error=None)
            except UserError as e:
                raise TerminalError(str(e)) from e

        id = ctx.tool_call_id
        if id is None:
            raise TerminalError("Tool call ID is required for turnstile synchronization.")
        context = current_context()
        if context is None:
            raise UserError(
                "A tool cannot be used without a Restate context. Make sure to run it within an agent or a run context."
            )
        turnstile = current_state().turnstile
        try:
            await turnstile.wait_for(id)
            res = await context.run_typed(f"Calling {name}", action, self.options)

            if res.kind == "call_deferred":
                raise CallDeferred()
            elif res.kind == "approval_required":
                raise ApprovalRequired()
            elif res.kind == "model_retry":
                assert res.error is not None
                raise ModelRetry(res.error)
            else:
                assert res.kind == "output"
                turnstile.allow_next_after(id)
                return res.output
        except Exception as e:
            turnstile.cancel_all_after(id)
            raise e from None
        except SdkInternalException as e:
            turnstile.cancel_all_after(id)
            raise RuntimeError() from e

    def visit_and_replace(
        self, visitor: Callable[[AbstractToolset[AgentDepsT]], AbstractToolset[AgentDepsT]]
    ) -> AbstractToolset[AgentDepsT]:
        return visitor(self)


class RestateMCPServer(WrapperToolset[AgentDepsT]):
    """A wrapper for MCPServer that integrates with restate."""

    def __init__(self, wrapped: MCPServer, run_options: RunOptions):
        super().__init__(wrapped)
        self._wrapped = wrapped
        self.get_tools_options = replace(run_options, serde=MCP_GET_TOOLS_SERDE)
        self.run_tools_options = replace(run_options, serde=MCP_RUN_SERDE)

    def visit_and_replace(
        self, visitor: Callable[[AbstractToolset[AgentDepsT]], AbstractToolset[AgentDepsT]]
    ) -> AbstractToolset[AgentDepsT]:
        return visitor(self)

    async def get_tools(self, ctx: RunContext[AgentDepsT]) -> dict[str, ToolsetTool[AgentDepsT]]:
        async def get_tools_in_context() -> RestateMCPGetToolsContextRunResult:
            res = await self._wrapped.get_tools(ctx)
            # ToolsetTool is not serializable as it holds a SchemaValidator
            # (which is also the same for every MCP tool so unnecessary to pass along the wire every time),
            # so we just return the ToolDefinitions and wrap them in ToolsetTool outside of the activity.
            return RestateMCPGetToolsContextRunResult(output={name: tool.tool_def for name, tool in res.items()})

        context = current_context()
        if context is None:
            raise UserError(
                "A toolset cannot be used without a Restate context. Make sure to run it within an agent or a run context."
            )

        try:
            tool_defs = await context.run_typed("get mcp tools", get_tools_in_context, self.get_tools_options)
            return {name: self.tool_for_tool_def(tool_def) for name, tool_def in tool_defs.output.items()}
        except SdkInternalBaseException as e:
            raise Exception("Internal error during get_tools call") from e

    def tool_for_tool_def(self, tool_def: ToolDefinition) -> ToolsetTool[AgentDepsT]:
        assert isinstance(self.wrapped, MCPServer)
        return self.wrapped.tool_for_tool_def(tool_def)

    async def call_tool(
        self,
        name: str,
        tool_args: dict[str, Any],
        ctx: RunContext[AgentDepsT],
        tool: ToolsetTool[AgentDepsT],
    ) -> ToolResult:
        async def call_tool_in_context() -> RestateMCPToolRunResult:
            res = await self._wrapped.call_tool(name, tool_args, ctx, tool)
            return RestateMCPToolRunResult(output=res)

        context = current_context()
        if context is None:
            raise UserError(
                "A toolset cannot be used without a Restate context. Make sure to run it within an agent or a run context."
            )
        try:
            res = await context.run_typed(f"Calling mcp tool {name}", call_tool_in_context, self.run_tools_options)
        except SdkInternalBaseException as e:
            raise Exception("Internal error during tool call") from e

        return res.output



#
#  Copyright (c) 2023-2025 - Restate Software, Inc., Restate GmbH
#
#  This file is part of the Restate SDK for Python,
#  which is released under the MIT license.
#
#  You can find a copy of the license in file LICENSE in the root
#  directory of this repository or package, or at
#  https://github.com/restatedev/sdk-typescript/blob/main/LICENSE
#
"""
This module contains the optional OpenAI integration for Restate.
"""

from agents.memory.session import SessionABC
from agents.items import TResponseInputItem
from typing import List, cast

from restate.extensions import current_context
from restate import ObjectContext


class RestateSession(SessionABC):
    """Restate session implementation following the Session protocol."""

    def __init__(self):
        self._items: List[TResponseInputItem] | None = None

    def _ctx(self) -> ObjectContext:
        return cast(ObjectContext, current_context())

    async def get_items(self, limit: int | None = None) -> List[TResponseInputItem]:
        """Retrieve conversation history for this session."""
        if self._items is None:
            self._items = await self._ctx().get("items") or []
        if limit is not None:
            return self._items[-limit:]
        return self._items.copy()

    async def add_items(self, items: List[TResponseInputItem]) -> None:
        """Store new items for this session."""
        if self._items is None:
            self._items = await self._ctx().get("items") or []
        self._items.extend(items)

    async def pop_item(self) -> TResponseInputItem | None:
        """Remove and return the most recent item from this session."""
        if self._items is None:
            self._items = await self._ctx().get("items") or []
        if self._items:
            return self._items.pop()
        return None

    def flush(self) -> None:
        """Flush the session items to the context."""
        self._ctx().set("items", self._items)

    async def clear_session(self) -> None:
        """Clear all items for this session."""
        self._items = []
        self._ctx().clear("items")
